import connexion

from fabric_cf.orchestrator.swagger_server.models.poa import Poa  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.poa_post import PoaPost  # noqa: E501
from fabric_cf.orchestrator.swagger_server.response import poas_controller as rc


def poas_create_sliver_id_post(body, sliver_id):  # noqa: E501
    """Perform an operational action on a sliver.

    Request to perform an operation action on a sliver. Supported actions include - reboot a VM sliver, get cpu info, get numa info, pin vCPUs, pin memory to a numa node etc.    # noqa: E501

    :param body: Perform Operation Action
    :type body: dict | bytes
    :param sliver_id: Sliver identified by universally unique identifier
    :type sliver_id: str

    :rtype: Poa
    """
    if connexion.request.is_json:
        body = PoaPost.from_dict(connexion.request.get_json())  # noqa: E501
    return rc.poas_create_sliver_id_post(body=body, sliver_id=sliver_id)


def poas_get(sliver_id=None, states=None, limit=None, offset=None):  # noqa: E501
    """Request get the status of the POAs.

    Request get the status of the POAs    # noqa: E501

    :param sliver_id: Search for POAs for a sliver
    :type sliver_id: str
    :param states: Search for POAs in the specified states
    :type states: List[str]
    :param limit: maximum number of results to return per page (1 or more)
    :type limit: int
    :param offset: number of items to skip before starting to collect the result set
    :type offset: int

    :rtype: Poa
    """
    return rc.poas_get(sliver_id=sliver_id, states=states, limit=limit, offset=offset)


def poas_poa_id_get(poa_id):  # noqa: E501
    """Perform an operational action on a sliver.

    Request get the status of the POA identified by poa_id.    # noqa: E501

    :param poa_id: Poa Id for the POA triggered
    :type poa_id: str

    :rtype: Poa
    """
    return rc.poas_poa_id_get(poa_id=poa_id)



from fim.graph.neo4j_property_graph import Neo4jGraphImporter, Neo4jPropertyGraph
from fim.graph.resources.neo4j_arm import Neo4jARMGraph
from fim.graph.resources.neo4j_cbm import Neo4jCBMGraph

import yaml
neo4j = None
with open("./config/config.test.yaml", 'r') as stream:
    try:
        config_dict = yaml.safe_load(stream)
        neo4j = config_dict["neo4j"]
    except yaml.YAMLError as exc:
        print(exc)


class FimTestHelper:
    n4j_imp = Neo4jGraphImporter(url=neo4j["url"], user=neo4j["user"],
                                 pswd=neo4j["pass"],
                                 import_host_dir=neo4j["import_host_dir"],
                                 import_dir=neo4j["import_dir"])

    @staticmethod
    def generate_adms() -> list:
        # these are produced by substrate tests
        site_ads = ['../../../neo4j/RENCI-ad.graphml', '../../../neo4j/UKY-ad.graphml',
                    '../../../neo4j/LBNL-ad.graphml', '../../../neo4j/Network-ad.graphml']

        result = []

        for ad in site_ads:
            plain_neo4j = FimTestHelper.n4j_imp.import_graph_from_file_direct(graph_file=ad)
            print(f"Validating ARM graph {ad}")
            plain_neo4j.validate_graph()

            site_arm = Neo4jARMGraph(graph=Neo4jPropertyGraph(graph_id=plain_neo4j.graph_id,
                                                              importer=FimTestHelper.n4j_imp))
            # generate a dict of ADMs from site graph ARM
            site_adms = site_arm.generate_adms()
            print('ADMS' + str(site_adms.keys()))

            # desired ADM is under 'primary'
            site_adm = site_adms['primary']
            result.append(site_adm)
        return result

    @staticmethod
    def generate_renci_adm():
        renci_ad = '../../../neo4j/RENCI-ad.graphml'
        cbm = Neo4jCBMGraph(importer=FimTestHelper.n4j_imp)
        plain_neo4j = FimTestHelper.n4j_imp.import_graph_from_file_direct(graph_file=renci_ad)
        print(f"Validating ARM graph {renci_ad}")
        plain_neo4j.validate_graph()

        site_arm = Neo4jARMGraph(graph=Neo4jPropertyGraph(graph_id=plain_neo4j.graph_id,
                                                          importer=FimTestHelper.n4j_imp))
        # generate a dict of ADMs from site graph ARM
        site_adms = site_arm.generate_adms()
        print('ADMS' + str(site_adms.keys()))

        # desired ADM is under 'primary'
        site_adm = site_adms['primary']

        return site_arm, site_adm



# coding: utf-8

from __future__ import absolute_import

from flask import json
from six import BytesIO

from fabric_cf.orchestrator.swagger_server.models.slivers import Slivers  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status400_bad_request import Status400BadRequest  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status401_unauthorized import Status401Unauthorized  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status403_forbidden import Status403Forbidden  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status404_not_found import Status404NotFound  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status500_internal_server_error import Status500InternalServerError  # noqa: E501
from fabric_cf.orchestrator.swagger_server.test import BaseTestCase


class TestSliversController(BaseTestCase):
    """SliversController integration test stubs"""

    def test_slivers_get(self):
        """Test case for slivers_get

        Retrieve a listing of user slivers
        """
        query_string = [('slice_id', 'slice_id_example'),
                        ('as_self', true)]
        response = self.client.open(
            '//slivers',
            method='GET',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slivers_sliver_id_get(self):
        """Test case for slivers_sliver_id_get

        slivers properties
        """
        query_string = [('slice_id', 'slice_id_example'),
                        ('as_self', true)]
        response = self.client.open(
            '//slivers/{sliver_id}'.format(sliver_id='sliver_id_example'),
            method='GET',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))


if __name__ == '__main__':
    import unittest
    unittest.main()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from datetime import datetime
from typing import List

from fabric_mb.message_bus.messages.poa_avro import PoaAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.reservation_predecessor_avro import ReservationPredecessorAvro
from fabric_mb.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.broker_query_model_avro import BrokerQueryModelAvro
from fabric_mb.message_bus.messages.get_actors_request_avro import GetActorsRequestAvro
from fabric_mb.message_bus.messages.proxy_avro import ProxyAvro
from fabric_mb.message_bus.messages.get_reservation_units_request_avro import GetReservationUnitsRequestAvro
from fabric_mb.message_bus.messages.unit_avro import UnitAvro
from fim.slivers.base_sliver import BaseSliver
from fim.user import GraphFormat

from fabric_cf.actor.core.apis.abc_actor_mixin import ActorType
from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.apis.abc_mgmt_controller_mixin import ABCMgmtControllerMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.manage.kafka.kafka_actor import KafkaActor
from fabric_cf.actor.core.util.id import ID


class KafkaController(KafkaActor, ABCMgmtControllerMixin):
    def clone(self):
        return KafkaController(guid=self.management_id,
                               kafka_topic=self.kafka_topic,
                               auth=self.auth, logger=self.logger,
                               message_processor=self.message_processor,
                               producer=self.producer)

    def add_broker(self, *, broker: ProxyAvro) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def update_broker(self, *, broker: ProxyAvro) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def get_brokers(self, *, broker: ID = None, id_token: str = None) -> List[ProxyAvro]:
        request = GetActorsRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token, broker_id=broker)
        request.type = ActorType.Broker.name
        status, response = self.send_request(request)

        if status.code == 0:
            return response.proxies
        return None

    def get_broker_query_model(self, *, broker: ID, id_token: str, level: int,
                               graph_format: GraphFormat) -> BrokerQueryModelAvro:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def claim_delegations(self, *, broker: ID, did: ID) -> DelegationAvro:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def reclaim_delegations(self, *, broker: ID, did: ID) -> DelegationAvro:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def get_reservation_units(self, *, rid: ID, id_token: str = None) -> List[UnitAvro]:
        request = GetReservationUnitsRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token, rid=rid)
        status, response = self.send_request(request)

        if status.code == 0:
            return response.units
        return None

    def add_reservation(self, *, reservation: TicketReservationAvro) -> ID:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def add_reservations(self, *, reservations: List[ReservationMng]) ->list:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def demand_reservation(self, *, reservation: ReservationMng) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def demand_reservation_rid(self, *, rid: ID) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def extend_reservation(self, *, reservation: ID, new_end_time: datetime, sliver: BaseSliver,
                           dependencies: List[ReservationPredecessorAvro] = None) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def modify_reservation(self, *, rid: ID, modify_properties: dict) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def poa(self, *, poa: PoaAvro) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from typing import List, Dict

from fabric_cf.actor.core.common.constants import Constants


class GlobalConfig:
    """
    Represent Global configuration
    """
    def __init__(self, *, config: dict):
        self.runtime = {}
        if Constants.CONFIG_SECTION_RUNTIME in config:
            self.runtime = config.get(Constants.CONFIG_SECTION_RUNTIME)

        self.logging = {}
        if Constants.CONFIG_LOGGING_SECTION in config:
            self.logging = config.get(Constants.CONFIG_LOGGING_SECTION)

        self.oauth = {}
        if Constants.CONFIG_SECTION_O_AUTH in config:
            self.oauth = config.get(Constants.CONFIG_SECTION_O_AUTH)

        self.database = {}
        if Constants.CONFIG_SECTION_DATABASE in config:
            self.database = config.get(Constants.CONFIG_SECTION_DATABASE)

        self.container = {}
        if Constants.CONFIG_SECTION_CONTAINER in config:
            self.container = config.get(Constants.CONFIG_SECTION_CONTAINER)

        self.time = {}
        if Constants.CONFIG_SECTION_TIME in config:
            self.time = config.get(Constants.CONFIG_SECTION_TIME)

        self.neo4j = {}
        if Constants.CONFIG_SECTION_NEO4J in config:
            self.neo4j = config[Constants.CONFIG_SECTION_NEO4J]

        self.pdp = {}
        if Constants.CONFIG_SECTION_PDP in config:
            self.pdp = config[Constants.CONFIG_SECTION_PDP]

        self.bqm = {}
        if Constants.CONFIG_SECTION_BQM in config:
            self.bqm = config[Constants.CONFIG_SECTION_BQM]

    def get_runtime(self) -> dict:
        """
        Return runtime config
        """
        return self.runtime

    def get_logging(self) -> dict:
        """
        Return logging config
        """
        return self.logging

    def get_oauth(self) -> dict:
        """
        Return oauth config
        """
        return self.oauth

    def get_database(self) -> dict:
        """
        Return database config
        """
        return self.database

    def get_container(self) -> dict:
        """
        Return container config
        """
        return self.container

    def get_time(self) -> dict:
        """
        Return time config
        """
        return self.time

    def get_neo4j_config(self) -> dict:
        """
        Return neo4j config
        """
        return self.neo4j

    def get_pdp_config(self) -> dict:
        """
        Return PDP config
        """
        return self.pdp

    def get_bqm_config(self) -> dict:
        """
        Return BQM config
        """
        return self.bqm


class HandlerConfig:
    """
    Represents Handler Config
    """
    def __init__(self, *, config: dict):
        self.module_name = config.get(Constants.PROPERTY_CONF_MODULE_NAME, None)
        self.class_name = config.get(Constants.PROPERTY_CONF_CLASS_NAME, None)
        self.properties = {}
        if Constants.PROPERTY_CONF_PROPERTIES_NAME in config:
            self.properties = config.get(Constants.PROPERTY_CONF_PROPERTIES_NAME)

    def get_module_name(self) -> str:
        """
        Return Module Name for Handler
        """
        return self.module_name

    def get_class_name(self) -> str:
        """
        Return Class Name for Handler
        """
        return self.class_name

    def get_properties(self) -> dict:
        """
        Return Handler Properties
        """
        return self.properties


class ControlConfig:
    """
    Represents Control or Inventory config
    """
    def __init__(self, *, config: dict):
        self.type = []
        self.module_name = config.get(Constants.PROPERTY_CONF_MODULE_NAME, None)
        self.class_name = config.get(Constants.PROPERTY_CONF_CLASS_NAME, None)
        if Constants.TYPE in config:
            value = config.get(Constants.TYPE)
            values = value.split(',')
            for v in values:
                self.type.append(v.strip())

    def get_type(self) -> List[str]:
        """
        Return type
        """
        return self.type

    def get_module_name(self) -> str:
        """
        Return Module name
        """
        return self.module_name

    def get_class_name(self) -> str:
        """
        Return Class Name
        """
        return self.class_name


class ResourceConfig:
    """
    Represents resource config
    """
    def __init__(self, *, config: dict):
        self.description = config.get(Constants.DESCRIPTION, None)
        self.type = []
        if Constants.TYPE in config:
            value = config.get(Constants.TYPE)
            values = value.split(',')
            for v in values:
                self.type.append(v.strip())
        self.label = config.get(Constants.LABEL, None)
        self.handler = HandlerConfig(config=config.get(Constants.HANDLER, None))

    def get_type(self) -> List[str]:
        """
        Return type
        """
        return self.type

    def get_label(self) -> str:
        """
        Return label
        """
        return self.label

    def get_description(self) -> str:
        """
        Return description
        """
        return self.description

    def get_handler(self) -> HandlerConfig:
        """
        Return handler config
        """
        return self.handler


class PolicyConfig:
    """
    Represents Policy config
    """
    def __init__(self, *, config: dict):
        self.module_name = config.get(Constants.PROPERTY_CONF_MODULE_NAME, None)
        self.class_name = config.get(Constants.PROPERTY_CONF_CLASS_NAME, None)
        self.properties = {}
        if Constants.PROPERTY_CONF_PROPERTIES_NAME in config:
            self.properties = config.get(Constants.PROPERTY_CONF_PROPERTIES_NAME)

    def get_module_name(self) -> str:
        """
        Return module name
        """
        return self.module_name

    def get_class_name(self) -> str:
        """
        Return class name
        """
        return self.class_name

    def get_properties(self) -> dict:
        """
        Return properties
        """
        return self.properties


class ActorConfig:
    """
    Represents Actor Config
    """
    def __init__(self, *, config: dict):
        self.policy = PolicyConfig(config=config.get('policy', None))
        self.description = config.get(Constants.DESCRIPTION, None)
        self.resources = []
        self.controls = []
        self.type = config.get(Constants.TYPE, None)
        self.name = config.get(Constants.NAME, None)
        self.guid = config.get(Constants.GUID, None)
        self.kafka_topic = config.get(Constants.KAFKA_TOPIC, None)
        self.substrate_file = config.get(Constants.PROPERTY_SUBSTRATE_FILE)
        res = config.get('resources', None)
        if res is not None:
            for r in res:
                resource_config = ResourceConfig(config=r['resource'])
                self.resources.append(resource_config)

        ctrls = config.get('controls', None)
        if ctrls is not None:
            for c in ctrls:
                control_config = ControlConfig(config=c['control'])
                self.controls.append(control_config)

    def get_type(self) -> str:
        """
        Return Actor Type
        """
        return self.type

    def get_name(self) -> str:
        """
        Return Actor Name
        """
        return self.name

    def get_guid(self) -> str:
        """
        Return Actor Guid
        """
        return self.guid

    def get_description(self) -> str:
        """
        Return Actor Description
        """
        return self.description

    def get_kafka_topic(self) -> str:
        """
        Return Kafka topic
        """
        return self.kafka_topic

    def get_controls(self) -> List[ControlConfig]:
        """
        Return Controls
        """
        return self.controls

    def get_policy(self) -> PolicyConfig:
        """
        Return Policy
        """
        return self.policy

    def get_resources(self) -> List[ResourceConfig]:
        """
        Return resources
        """
        return self.resources

    def get_substrate_file(self) -> str:
        """
        Return Substrate file
        """
        return self.substrate_file


class Peer:
    """
    Represent Peer Config
    """
    def __init__(self, *, config: dict):
        self.name = config.get(Constants.NAME, None)
        self.type = config.get(Constants.TYPE, None)
        self.guid = config.get(Constants.GUID, None)
        self.kafka_topic = config.get(Constants.KAFKA_TOPIC, None)
        self.delegation = config.get(Constants.DELEGATION, None)

    def get_name(self) -> str:
        """
        Return Name
        """
        return self.name

    def get_type(self) -> str:
        """
        Return Type
        """
        return self.type

    def get_guid(self) -> str:
        """
        Return Guid
        """
        return self.guid

    def get_kafka_topic(self) -> str:
        """
        Return Kafka Topic
        """
        return self.kafka_topic

    def get_delegation(self) -> str:
        """
        Return delegation
        """
        return self.delegation


class Configuration:
    """
    Represent Configuration read from file
    """
    def __init__(self, *, config: dict):
        self.global_config = GlobalConfig(config=config)
        self.actor = ActorConfig(config=config['actor'])
        self.peers = []
        self.topic_peer_map = {}
        self.peers_dict = {}
        if 'peers' in config:
            for e in config['peers']:
                p = Peer(config=e['peer'])
                self.peers.append(p)
                self.topic_peer_map[p.get_kafka_topic()] = p
                self.peers_dict[p.get_name()] = p

    def get_global_config(self) -> GlobalConfig:
        """
        Return Global Config
        """
        return self.global_config

    def get_log_config(self) -> dict:
        """
        Return Log config
        """
        return self.global_config.get_logging()

    def get_runtime_config(self) -> dict:
        """
        Return Runtime Config
        """
        if self.global_config is not None:
            return self.global_config.get_runtime()
        return None

    def get_oauth_config(self) -> dict:
        """
        Return OAuth Config
        """
        if self.global_config is not None:
            return self.global_config.get_oauth()
        return None

    def get_actor_config(self) -> ActorConfig:
        """
        Return Actor Config
        """
        return self.actor

    def get_peers(self) -> List[Peer]:
        """
        Return Peer Config
        """
        return self.peers

    def get_peers_dict(self) -> Dict[str, Peer]:
        """
        Return Peer Config
        """
        return self.peers_dict

    def get_topic_peer_map(self) -> Dict[str, Peer]:
        return self.topic_peer_map

    def get_neo4j_config(self) -> dict:
        """
        Return Neo4j config
        """
        if self.global_config is not None:
            return self.global_config.get_neo4j_config()

        return None

    def get_rpc_retries(self) -> int:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_RPC_RETRIES, 5)

    def get_kafka_key_schema_location(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_KEY_SCHEMA, None)

    def get_kafka_value_schema_location(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_VALUE_SCHEMA, None)

    def get_kafka_server(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SERVER, None)

    def get_kafka_schema_registry(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SCHEMA_REGISTRY, None)

    def get_kafka_security_protocol(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SECURITY_PROTOCOL, None)

    def get_kafka_ssl_ca_location(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_S_SL_CA_LOCATION, None)

    def get_kafka_ssl_cert_location(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SSL_CERTIFICATE_LOCATION, None)

    def get_kafka_ssl_key_location(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SSL_KEY_LOCATION, None)

    def get_kafka_ssl_key_password(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SSL_KEY_PASSWORD, None)

    def get_kafka_prod_user_name(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_USERNAME, None)

    def get_kafka_prod_user_pwd(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_PASSWORD, None)

    def get_kafka_sasl_mechanism(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SASL_MECHANISM, None)

    def get_kafka_cons_group_id(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_GROUP_ID, None)

    def get_kafka_cons_user_name(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SASL_CONSUMER_USERNAME, None)

    def get_kafka_cons_user_pwd(self) -> str or None:
        return self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_SASL_CONSUMER_PASSWORD, None)

    def get_kafka_request_timeout_ms(self) -> int:
        value = self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_REQUEST_TIMEOUT_MS, 120000)
        return int(value)

    def get_kafka_max_message_size(self) -> int:
        value = self.global_config.runtime.get(Constants.PROPERTY_CONF_KAFKA_MAX_MESSAGE_SIZE, 1048588)
        return int(value)

    def get_rpc_request_timeout_seconds(self) -> int:
        value = self.global_config.runtime.get(Constants.PROPERTY_CONF_RPC_REQUEST_TIMEOUT_SECONDS, 900)
        return int(value)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
    from fabric_cf.actor.core.apis.abc_response_handler import ABCResponseHandler
    from fabric_cf.actor.core.kernel.rpc_request_type import RPCRequestType
    from fabric_cf.actor.security.auth_token import AuthToken


class IncomingRPC:
    """
    Represents an incoming RPC message
    """
    def __init__(self, *, message_id: ID, request_type: RPCRequestType,
                 callback: ABCCallbackProxy, caller: AuthToken):
        self.request_type = request_type
        self.message_id = message_id
        self.callback = callback
        self.caller = caller
        self.response_handler = None
        self.error = None
        self.request_id = None

    def get_request_type(self) -> RPCRequestType:
        """
        Get request type
        @return request type
        """
        return self.request_type

    def get_callback(self) -> ABCCallbackProxy:
        """
        Get callback
        @return callback
        """
        return self.callback

    def get_caller(self) -> AuthToken:
        """
        Get caller
        @return caller
        """
        return self.caller

    def get_message_id(self) -> ID:
        """
        Get message_id
        @return message_id
        """
        return self.message_id

    def set_response_handler(self, *, response_handler: ABCResponseHandler):
        """
        Set response_handler
        @param response_handler response_handler
        """
        self.response_handler = response_handler

    def get_response_handler(self) -> ABCResponseHandler:
        """
        Get response_handler
        @return response_handler
        """
        return self.response_handler

    def set_error(self, *, error: Exception):
        """
        Set error
        @param error error
        """
        self.error = error

    def get_error(self) -> Exception:
        """
        Get error
        @return error
        """
        return self.error

    def set_request_id(self, *, request_id: ID):
        """
        Set request id
        @param request_id request id
        """
        self.request_id = request_id

    def get_request_id(self) -> ID:
        """
        Get request_id
        @return request_id
        """
        return self.request_id

    def __str__(self):
        return "MessageID={} requestType={} caller={}:{}".format(self.message_id, self.request_type,
                                                                 self.caller.get_name(), self.caller.get_guid())

    @abstractmethod
    def get(self):
        """
        Return delegation or reservation or dictionary containing the request
        """

    @abstractmethod
    def get_update_data(self):
        """
        Return updated data
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading

from fim.graph.abc_property_graph import ABCPropertyGraph

from fabric_cf.actor.fim.fim_helper import FimHelper
from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin, ActorType
from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation, DelegationState
from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import DelegationException
from fabric_cf.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.update_data import UpdateData
from fabric_cf.actor.security.auth_token import AuthToken


class Delegation(ABCDelegation):
    error_string_prefix = 'error for delegation: {} : {}'
    invalid_state_prefix = "Invalid state for {}. Did you already {} this Delegation?"

    def __init__(self, dlg_graph_id: str, slice_id: ID, delegation_name: str = None, site: str = None):
        self.dlg_graph_id = dlg_graph_id
        self.slice_id = slice_id
        self.state = DelegationState.Nascent
        self.dirty = False
        self.state_transition = False
        self.sequence_in = 0
        self.update_count = 0
        self.sequence_out = 0
        self.update_data = UpdateData()
        self.must_send_update = False
        self.graph = None
        self.actor = None
        self.slice_object = None
        self.logger = None
        self.policy = None
        self.callback = None
        self.error_message = None
        self.owner = None
        self.delegation_name = delegation_name
        self.site = site
        self.thread_lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['graph']
        del state['actor']
        del state['slice_object']
        del state['logger']
        del state['policy']
        del state['thread_lock']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.graph = None
        self.actor = None
        self.slice_object = None
        self.logger = None
        self.policy = None
        self.thread_lock = threading.Lock()

    def restore(self, actor: ABCActorMixin, slice_obj: ABCSlice):
        self.actor = actor
        self.slice_object = slice_obj
        if actor is not None:
            self.logger = actor.get_logger()
            if self.callback is not None:
                self.callback.set_logger(logger=actor.get_logger())
        if actor.get_type() == ActorType.Authority:
            self.graph = FimHelper.get_arm_graph(graph_id=str(self.dlg_graph_id))

        if actor.get_policy() is not None:
            self.policy = actor.get_policy()

    def get_site(self) -> str:
        return self.site

    def set_graph(self, graph: ABCPropertyGraph):
        self.graph = graph

    def get_graph(self) -> ABCPropertyGraph:
        return self.graph

    def get_actor(self) -> ABCActorMixin:
        return self.actor

    def get_delegation_id(self) -> str:
        return self.dlg_graph_id

    def get_slice_id(self) -> ID:
        if self.slice_id is not None:
            return self.slice_id
        elif self.slice_object is not None:
            return self.slice_object.get_slice_id()

    def get_slice_object(self) -> ABCSlice:
        return self.slice_object

    def get_state(self) -> DelegationState:
        return self.state

    def get_state_name(self) -> str:
        return self.state.name

    def set_logger(self, *, logger):
        self.logger = logger

    def set_slice_object(self, *, slice_object: ABCSlice):
        self.slice_object = slice_object

    def transition(self, *, prefix: str, state: DelegationState):
        if self.logger is not None:
            self.logger.debug("Delegation #{} {} transition: {} -> {}".format(self.dlg_graph_id, prefix,
                                                                              self.get_state_name(),
                                                                              state.name))

        self.state = state
        self.set_dirty()
        self.state_transition = True

    def clear_notice(self):
        """
        Clear the notices
        """
        self.update_data.clear()

    def get_notices(self) -> str:
        s = "Delegation {} (Slice {} ) is in state [{}]".format(self.get_delegation_id(), self.get_slice_id(),
                                                                DelegationState(self.state).name)
        if self.error_message is not None and self.error_message != "":
            s += ", err={}".format(self.error_message)

        notices = self.update_data.get_events()
        if notices is not None:
            s += "\n{}".format(notices)

        notices = self.update_data.get_message()
        if notices is not None:
            s += "\n{}".format(notices)

        return s

    def has_uncommitted_transition(self) -> bool:
        return self.state_transition

    def is_dirty(self) -> bool:
        return self.dirty

    def set_dirty(self):
        self.dirty = True

    def clear_dirty(self):
        self.dirty = False

    def set_actor(self, actor: ABCActorMixin):
        self.actor = actor

    def prepare(self, *, callback: ABCCallbackProxy, logger):
        """
        Prepare a delegation
        """
        self.set_logger(logger=logger)
        self.callback = callback

        # Null callback indicates a locally initiated request to create an
        # exported delegation. Else the request is from a client and must have
        # a client-specified delegation id.

        if self.callback is not None and self.dlg_graph_id is None:
            self.logger.error(self.error_string_prefix.format(self, Constants.NOT_SPECIFIED_PREFIX.format("graph id")))
            raise DelegationException(Constants.NOT_SPECIFIED_PREFIX.format("graph id"))

        self.set_dirty()

    def is_closed(self) -> bool:
        return self.state == DelegationState.Closed

    def is_delegated(self) -> bool:
        return self.state == DelegationState.Delegated

    def is_reclaimed(self) -> bool:
        return self.state == DelegationState.Reclaimed

    def is_failed(self) -> bool:
        return self.state == DelegationState.Failed

    def delegate(self, policy: ABCPolicy):
        # These handlers may need to be slightly more sophisticated, since a
        # client may bid multiple times on a ticket as part of an auction
        # protocol: so we may receive a reserve or extend when there is already
        # a request pending.
        self.incoming_request()

        self.policy = policy
        self.map_and_update(delegated=False)

    def claim(self):
        if self.state == DelegationState.Delegated:
            # We are an agent asked to return a pre-reserved "will call" ticket
            # to a client. Set mustSendUpdate so that the update will be sent
            # on the next probe.
            self.must_send_update = True
        elif self.state == DelegationState.Reclaimed:
            self.transition(prefix="claim", state=DelegationState.Delegated)
            self.must_send_update = True
        else:
            self.logger.error(self.error_string_prefix.format(self, self.invalid_state_prefix.format('claim', 'claim')))
            raise DelegationException(self.invalid_state_prefix.format('claim', 'claim'))

    def reclaim(self):
        if self.state == DelegationState.Delegated:
            self.policy.reclaim(delegation=self)
            self.transition(prefix="reclaimed", state=DelegationState.Reclaimed)
        else:
            self.logger.error(self.error_string_prefix.format(self, self.invalid_state_prefix.format('reclaim',
                                                                                                     'reclaim')))
            raise DelegationException(self.invalid_state_prefix.format('reclaim', 'reclaim'))

    def close(self):
        send_notification = False
        if self.state == DelegationState.Nascent:
            self.logger.warning("Closing a delegation in progress")
            send_notification = True

        if self.state != DelegationState.Closed:
            self.transition(prefix="closed", state=DelegationState.Closed)
            self.policy.closed_delegation(delegation=self)

        if send_notification:
            self.update_data.error(message="Closed while advertising delegation")
            self.generate_update()

    def incoming_request(self):
        """
        Checks reservation state prior to handling an incoming request. These
        checks are not applied to probes or closes.

        @throws Exception
        """
        assert self.slice_object is not None

        # Disallow any further requests on a closed delegation. Generate and update to reset the client.
        if self.is_closed():
            self.generate_update()
            self.logger.error(self.error_string_prefix.format(self, "server cannot satisfy request closing"))
            raise DelegationException("server cannot satisfy request closing")

    def generate_update(self):
        """
        Generate an update
        """
        self.logger.debug("Generating update")
        if self.callback is None:
            self.logger.warning("Cannot generate update: no callback.")
            return

        self.logger.debug("Generating update: update count={}".format(self.update_count))
        try:
            self.update_count += 1
            self.sequence_out += 1
            RPCManagerSingleton.get().update_delegation(delegation=self)
            self.must_send_update = False
        except Exception as e:
            # Note that this may result in a "stuck" reservation... not much we
            # can do if the receiver has failed or rejects our update. We will
            # regenerate on any user-initiated probe.
            self.logger.error("callback failed, exception={}".format(e))

    def map_and_update(self, *, delegated: bool):
        """
        Call the policy to fill a request, with associated state transitions.
        Catch exceptions and report all errors using callback mechanism.

        @param delegated
                   true if this is delegated (i.e., request is reclaim)
        @return boolean success
        """
        success = False
        granted = False

        if self.state == DelegationState.Nascent:
            if delegated:
                self.fail_notify(message="delegation is not yet delegated")
            else:
                self.logger.debug("Using policy {} to bind delegation".format(self.policy.__class__.__name__))
                try:
                    granted = False
                    # If the policy has processed this reservation, granted should
                    # be set true so that we can send the result back to the
                    # client. If the policy has not yet processed this reservation
                    # (binPending is true) then call the policy. The policy may
                    # choose to process the request immediately (true) or to defer
                    # it (false). In case of a deferred request, we will eventually
                    # come back to this method after the policy has done its job.
                    granted = self.policy.bind_delegation(delegation=self)

                except Exception as e:
                    self.logger.error("map_and_update bind_delegation failed for advertise: {e}")
                    self.fail_notify(message=str(e))
                    return success

                if granted:
                    self.logger.debug("Delegation {} has been granted".format(self.get_delegation_id()))
                    success = True
                    self.transition(prefix="delegated", state=DelegationState.Delegated)
        elif self.state == DelegationState.Delegated:
            if not delegated:
                self.fail_notify(message="delegation is already ticketed")
        else:
            self.logger.error("map_and_update: unexpected state")
            self.fail_notify(message="invalid operation for the current reservation state")

        return success

    def fail_notify(self, *, message: str):
        """
        Notify a failed delegation by generating an update
        @param message message
        """
        self.error_message = message
        self.generate_update()
        self.logger.error(message)

    def service_delegate(self):
        if self.dlg_graph_id is not None:
            # Update the graph
            self.transition(prefix="update absorbed", state=DelegationState.Delegated)
            self.generate_update()

    def prepare_probe(self):
        return

    def probe_pending(self):
        return

    def service_probe(self):
        if self.must_send_update:
            self.generate_update()

    def __str__(self):
        msg = "del: "
        if self.dlg_graph_id is not None:
            msg += "#{} ".format(self.dlg_graph_id)

        if self.site is not None:
            msg += "site: {} ".format(self.site)

        if self.slice_object is not None:
            msg += "slice: [{}] ".format(self.slice_object)
        elif self.slice_id is not None:
            msg += "slice_id: [{}] ".format(self.get_slice_id())

        msg += "state:[{}] ".format(self.get_state_name())

        if self.graph is not None:
            msg += "graph:[{}] ".format(self.graph)

        return msg

    def get_callback(self) -> ABCCallbackProxy:
        return self.callback

    def get_update_data(self) -> UpdateData:
        return self.update_data

    def set_sequence_in(self, value: int):
        """
        Set incoming sequence number
        @param value value
        """
        self.sequence_in = value

    def get_sequence_in(self) -> int:
        return self.sequence_in

    def get_sequence_out(self) -> int:
        return self.sequence_out

    def update_delegation(self, *, incoming: ABCDelegation, update_data: UpdateData):
        self.logger.error(self.error_string_prefix.format(self, "Cannot update a authority delegation"))
        raise DelegationException("Cannot update a authority delegation")

    def service_update_delegation(self):
        return

    def validate(self):
        if self.slice_object is None:
            self.logger.error(self.error_string_prefix.format(self, Constants.NOT_SPECIFIED_PREFIX.format("slice")))
            raise DelegationException(Constants.NOT_SPECIFIED_PREFIX.format("slice"))

        if self.dlg_graph_id is None:
            self.logger.error(self.error_string_prefix.format(self, Constants.NOT_SPECIFIED_PREFIX.format("graph id")))
            raise DelegationException(Constants.NOT_SPECIFIED_PREFIX.format("graph id"))

    def validate_incoming(self):
        self.validate()

    def validate_outgoing(self):
        self.validate()

    def set_owner(self, *, owner: AuthToken):
        """
        Set owner
        @param owner owner
        """
        self.owner = owner

    def load_graph(self, *, graph_str: str):
        self.graph = FimHelper.get_graph_from_string_direct(graph_str=graph_str)

    def get_delegation_name(self) -> str:
        return self.delegation_name

    def fail(self, *, message: str, exception: Exception = None):
        """
        Fail a delegation
        """
        self.error_message = message
        self.update_data.error(message=message)
        self.transition(prefix=message, state=DelegationState.Failed)
        self.logger.error(f"{message}  e: {exception}")

    def lock(self):
        self.thread_lock.acquire()

    def unlock(self):
        if self.thread_lock.locked():
            self.thread_lock.release()


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING

from fabric_mb.message_bus.messages.failed_rpc_avro import FailedRpcAvro
from fabric_mb.message_bus.messages.query_avro import QueryAvro
from fabric_mb.message_bus.messages.query_result_avro import QueryResultAvro
from fabric_mb.message_bus.producer import AvroProducerApi

from fabric_cf.actor.core.common.exceptions import ProxyException
from fabric_cf.actor.core.proxies.kafka.translate import Translate
from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.rpc_request_state import RPCRequestState
from fabric_cf.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric_cf.actor.core.proxies.proxy import Proxy

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_rpc_request_state import ABCRPCRequestState
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.util.id import ID


class KafkaProxyRequestState(RPCRequestState):
    def __init__(self):
        super().__init__()
        self.callback_topic = None
        self.reservation = None
        self.delegation = None
        self.udd = None
        self.query = None
        self.poa = None
        self.request_id = None
        self.failed_reservation_id = None
        self.failed_request_type = None
        self.error_detail = None


class KafkaProxy(Proxy, ABCCallbackProxy):
    TypeDefault = 0
    TypeReturn = 1
    TypeBroker = 2
    TypeSite = 3

    def __init__(self, *, kafka_topic: str, identity: AuthToken, logger):
        super().__init__(auth=identity)
        self.kafka_topic = kafka_topic
        self.logger = logger
        self.proxy_type = Constants.PROTOCOL_KAFKA
        self.type = self.TypeDefault

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        peer_dict = GlobalsSingleton.get().get_config().get_peers_dict()
        self.logger = GlobalsSingleton.get().get_logger()
        peer = peer_dict.get(self.get_name())
        if peer is not None:
            self.kafka_topic = peer.get_kafka_topic()
        self.logger = None

    def execute(self, *, request: KafkaProxyRequestState, producer: AvroProducerApi):
        avro_message = None
        if request.get_type() == RPCRequestType.Query:
            avro_message = QueryAvro()
            avro_message.message_id = str(request.get_message_id())
            avro_message.properties = request.query
            avro_message.callback_topic = request.callback_topic
            avro_message.auth = Translate.translate_auth_to_avro(auth=request.caller)

        elif request.get_type() == RPCRequestType.QueryResult:
            avro_message = QueryResultAvro()
            avro_message.message_id = str(request.get_message_id())
            avro_message.request_id = str(request.request_id)
            avro_message.properties = request.query
            avro_message.auth = Translate.translate_auth_to_avro(auth=request.caller)

        elif request.get_type() == RPCRequestType.FailedRPC:
            avro_message = FailedRpcAvro()
            avro_message.message_id = str(request.get_message_id())
            avro_message.request_id = str(request.request_id)
            avro_message.request_type = request.failed_request_type.value
            avro_message.auth = Translate.translate_auth_to_avro(auth=request.caller)

            if request.failed_reservation_id is not None:
                avro_message.reservation_id = request.failed_reservation_id
            else:
                avro_message.reservation_id = ""
            avro_message.error_details = request.error_detail

        else:
            raise ProxyException("Unsupported RPC: type={}".format(request.get_type()))

        if producer is not None and producer.produce(topic=self.kafka_topic, record=avro_message):
            self.logger.debug("Message {} written to {}".format(avro_message.name, self.kafka_topic))
        else:
            self.logger.error("Failed to send message {} to {} via producer {}".format(avro_message.name,
                                                                                       self.kafka_topic, producer))

    def prepare_query(self, *, callback: ABCCallbackProxy, query: dict, caller: AuthToken):
        request = KafkaProxyRequestState()
        request.query = query
        request.callback_topic = callback.get_kafka_topic()
        request.caller = caller
        return request

    def prepare_query_result(self, *, request_id: str, response, caller: AuthToken) -> ABCRPCRequestState:
        request = KafkaProxyRequestState()
        request.query = response
        request.request_id = request_id
        request.caller = caller
        return request

    def prepare_failed_request(self, *, request_id: str, failed_request_type,
                               failed_reservation_id: ID, error: str, caller: AuthToken) -> ABCRPCRequestState:
        request = KafkaProxyRequestState()
        request.failed_request_type = failed_request_type
        request.failed_reservation_id = failed_reservation_id
        request.error_detail = error
        request.request_id = request_id
        request.caller = caller
        return request

    def get_kafka_topic(self):
        return self.kafka_topic



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import argparse
import logging
import traceback
from datetime import datetime, timezone, timedelta
from logging.handlers import RotatingFileHandler

import yaml
from fabric_mb.message_bus.messages.auth_avro import AuthAvro
from fim.slivers.network_service import ServiceType

from fabric_cf.actor.core.apis.abc_actor_mixin import ActorType
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates
from fabric_cf.actor.core.kernel.slice_state_machine import SliceState
from fabric_cf.actor.core.manage.kafka.kafka_actor import KafkaActor
from fabric_cf.actor.core.manage.kafka.kafka_mgmt_message_processor import KafkaMgmtMessageProcessor
from fabric_cf.actor.core.plugins.db.actor_database import ActorDatabase
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.fim.fim_helper import FimHelper


class MainClass:
    """
    Audit CLI provides following capabilities:
    - Remove/Delete slices older than specified number of days
    - Remove/Delete dangling network services which connect the ports to deleted/closed VMs
    """
    def __init__(self, config_file: str):
        with open(config_file) as f:
            config_dict = yaml.safe_load(f)

        # Load the config file
        self.log_config = config_dict[Constants.CONFIG_LOGGING_SECTION]

        # create the logger
        self.logger = logging.getLogger(self.log_config[Constants.PROPERTY_CONF_LOGGER])
        file_handler = RotatingFileHandler(f"{self.log_config[Constants.PROPERTY_CONF_LOG_DIRECTORY]}/audit.log",
                                           backupCount=self.log_config[Constants.PROPERTY_CONF_LOG_RETAIN],
                                           maxBytes=self.log_config[Constants.PROPERTY_CONF_LOG_SIZE])
        logging.basicConfig(level=logging.DEBUG,
                            format="%(asctime)s [%(filename)s:%(lineno)d] [%(levelname)s] %(message)s",
                            handlers=[logging.StreamHandler(), file_handler])

        # Neo4j Config
        self.neo4j_config = config_dict[Constants.CONFIG_SECTION_NEO4J]

        # Database Config
        self.database_config = config_dict[Constants.CONFIG_SECTION_DATABASE]

        # Actor Config
        self.actor_config = config_dict[Constants.CONFIG_SECTION_ACTOR]

        # Load config in the GlobalsSingleton
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        GlobalsSingleton.get().config_file = config_file
        GlobalsSingleton.get().load_config()
        GlobalsSingleton.get().initialized = True
        GlobalsSingleton.get().log = self.logger

        self.mgmt_actor = None

        # For Orchestrator; initialize a management actor
        actor_type = self.actor_config[Constants.TYPE]
        if actor_type.lower() == ActorType.Orchestrator.name.lower():
            actor_topic = self.actor_config[Constants.KAFKA_TOPIC]
            actor_topic = actor_topic.replace("^", "")
            actor_topic = actor_topic.replace("*", "1")
            actor_guid = GlobalsSingleton.get().get_config().get_actor_config().get_guid()

            self.audit_config = config_dict['audit']
            audit_topic = self.audit_config[Constants.KAFKA_TOPIC]

            # Setup Kafka producer config
            producer_config = GlobalsSingleton.get().get_kafka_config_producer()
            if Constants.SASL_USERNAME in producer_config:
                producer_config[Constants.SASL_USERNAME] = self.audit_config[Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_USERNAME]
                producer_config[Constants.SASL_PASSWORD] = self.audit_config[Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_PASSWORD]
                producer_config[Constants.SASL_MECHANISM] = self.audit_config[Constants.PROPERTY_CONF_KAFKA_SASL_MECHANISM]

            key_schema_loc = GlobalsSingleton.get().get_config().get_kafka_key_schema_location()
            val_schema_loc = GlobalsSingleton.get().get_config().get_kafka_value_schema_location()

            # Create a producer
            from fabric_mb.message_bus.producer import AvroProducerApi
            producer = AvroProducerApi(producer_conf=producer_config, key_schema_location=key_schema_loc,
                                       value_schema_location=val_schema_loc, logger=self.logger)

            # Setup Kafka consumer config
            consumer_config = GlobalsSingleton.get().get_kafka_config_consumer()
            if Constants.SASL_USERNAME in consumer_config:
                consumer_config[Constants.SASL_USERNAME] = self.audit_config[Constants.PROPERTY_CONF_KAFKA_SASL_CONSUMER_USERNAME]
                consumer_config[Constants.SASL_PASSWORD] = self.audit_config[Constants.PROPERTY_CONF_KAFKA_SASL_CONSUMER_PASSWORD]
                consumer_config[Constants.SASL_MECHANISM] = self.audit_config[Constants.PROPERTY_CONF_KAFKA_SASL_MECHANISM]

            # Setup Kafka Message processor to handle responses back from the actor
            auth = AuthAvro()
            auth.name = self.audit_config[Constants.NAME]
            auth.guid = self.audit_config[Constants.GUID]

            message_processor = KafkaMgmtMessageProcessor(consumer_conf=consumer_config,
                                                          key_schema_location=key_schema_loc,
                                                          value_schema_location=val_schema_loc, logger=self.logger,
                                                          topics=[audit_topic])

            self.mgmt_actor = KafkaActor(guid=ID(uid=actor_guid), kafka_topic=actor_topic,
                                         auth=auth, logger=self.logger, producer=producer,
                                         message_processor=message_processor)
            self.mgmt_actor.callback_topic = audit_topic

    def delete_dangling_network_slivers(self):
        """
        Delete dangling network slivers connected to VMs in Closed state
        It has been observed that some times failure to renew VMs but successful renew of Network Services leaves
        dangling NS attached to ports on VMs in closed state. This results in future NS
        provisioning when the same port is now assigned to a new VM. This command identifies such NS and cleans them.
        """
        actor_type = self.actor_config[Constants.TYPE]
        if actor_type.lower() != ActorType.Orchestrator.name.lower():
            return

        actor_db = ActorDatabase(user=self.database_config[Constants.PROPERTY_CONF_DB_USER],
                                 password=self.database_config[Constants.PROPERTY_CONF_DB_PASSWORD],
                                 database=self.database_config[Constants.PROPERTY_CONF_DB_NAME],
                                 db_host=self.database_config[Constants.PROPERTY_CONF_DB_HOST],
                                 logger=self.logger)

        states = [ReservationStates.Active.value, ReservationStates.ActiveTicketed.value,
                  ReservationStates.Failed.value]
        resource_type = []
        for s in ServiceType:
            resource_type.append(str(s))

        # Get the Active NS Slivers
        slivers = actor_db.get_reservations(states=states, rsv_type=resource_type)
        for s in slivers:
            # Check dependencies
            closed_preds = 0
            for pred_state in s.get_redeem_predecessors():
                if pred_state.get_reservation().is_closed() or pred_state.get_reservation().is_closing():
                    closed_preds += 1

            # Close the sliver if at least one of the dependencies are in closed state
            if closed_preds:
                self.logger.info(f"Closing Sliver: {s.get_reservation_id()};"
                                 f" Found dependencies# {closed_preds} in closed/closing state")
                # Trigger close
                self.mgmt_actor.close_reservation(rid=s.get_reservation_id())

    def delete_dead_closing_slice(self, *, days: int):
        """
        Delete Dead/Closing slice older than days* from postgres and neo4j database
        @param days: Number of days
        """

        # Create Db object
        actor_db = ActorDatabase(user=self.database_config[Constants.PROPERTY_CONF_DB_USER],
                                 password=self.database_config[Constants.PROPERTY_CONF_DB_PASSWORD],
                                 database=self.database_config[Constants.PROPERTY_CONF_DB_NAME],
                                 db_host=self.database_config[Constants.PROPERTY_CONF_DB_HOST],
                                 logger=self.logger)

        # Get Closing/Dead slices older than days
        states = [SliceState.Dead.value, SliceState.Closing.value]
        lease_end = datetime.now(timezone.utc) - timedelta(days=days)
        slices = actor_db.get_slices(states=states, lease_end=lease_end)
        actor_type = self.actor_config[Constants.TYPE]
        for s in slices:
            try:
                # Delete the slice from Neo4J - only done on orchestrator
                if actor_type.lower() == ActorType.Orchestrator.name.lower():
                    try:
                        FimHelper.delete_graph(graph_id=s.get_graph_id(), neo4j_config=self.neo4j_config)
                    except Exception as e:
                        self.logger.error(f"Failed to delete graph {s.get_graph_id()} for Slice# {s.get_slice_id()}: e: {e}")
                        self.logger.error(traceback.format_exc())

                # Fetch the slivers and delete them from Postgres
                reservations = actor_db.get_reservations(slice_id=s.get_slice_id())
                for r in reservations:
                    try:
                        actor_db.remove_reservation(rid=r.get_reservation_id())
                    except Exception as e:
                        self.logger.error(
                            f"Failed to delete reservation {r.get_reservation_id()} for Slice# {s.get_slice_id()}: e: {e}")
                        self.logger.error(traceback.format_exc())

                # Remove the slice from Postgres
                actor_db.remove_slice(slice_id=s.get_slice_id())
            except Exception as e:
                self.logger.error(f"Failed to delete slice: {s.get_slice_id()}: e: {e}")
                self.logger.error(traceback.format_exc())

    def handle_command(self, args):
        """
        Command Handler
        @param args Command arguments
        """
        # Slice commands
        if args.command == "slices":
            # Remove operation
            if args.operation is not None and args.operation == "remove":
                self.delete_dead_closing_slice(days=args.days)
            else:
                print(f"Unsupported operation: {args.operation}")
        # Slivers
        elif args.command == "slivers":
            # Close operation
            if args.operation is not None and args.operation == "close":
                self.delete_dangling_network_slivers()
            else:
                print(f"Unsupported operation: {args.operation}")
        else:
            print(f"Unsupported command: {args.command}")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", dest='config', required=True, type=str)
    parser.add_argument("-d", dest='days', required=False, type=int, default=30)
    parser.add_argument("-c", dest='command', required=True, type=str)
    parser.add_argument("-o", dest='operation', required=True, type=str)
    args = parser.parse_args()

    mc = MainClass(config_file=args.config)
    mc.handle_command(args)




# coding: utf-8

from __future__ import absolute_import

from flask import json
from six import BytesIO

from fabric_cf.orchestrator.swagger_server.models.slice_details import SliceDetails  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.slices import Slices  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.slices_post import SlicesPost  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.slivers import Slivers  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status200_ok_no_content import Status200OkNoContent  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status400_bad_request import Status400BadRequest  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status401_unauthorized import Status401Unauthorized  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status403_forbidden import Status403Forbidden  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status404_not_found import Status404NotFound  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status500_internal_server_error import Status500InternalServerError  # noqa: E501
from fabric_cf.orchestrator.swagger_server.test import BaseTestCase


class TestSlicesController(BaseTestCase):
    """SlicesController integration test stubs"""

    def test_slices_create_post(self):
        """Test case for slices_create_post

        Create slice
        """
        body = 'body_example'
        query_string = [('name', 'name_example'),
                        ('ssh_key', 'ssh_key_example'),
                        ('lease_end_time', 'lease_end_time_example')]
        response = self.client.open(
            '//slices/create',
            method='POST',
            data=json.dumps(body),
            content_type='text/plain',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_creates_post(self):
        """Test case for slices_creates_post

        Create slice
        """
        body = SlicesPost()
        query_string = [('name', 'name_example'),
                        ('lease_end_time', 'lease_end_time_example')]
        response = self.client.open(
            '//slices/creates',
            method='POST',
            data=json.dumps(body),
            content_type='application/json',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_delete_delete(self):
        """Test case for slices_delete_delete

        Delete all slices for a User within a project.
        """
        response = self.client.open(
            '//slices/delete',
            method='DELETE')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_delete_slice_id_delete(self):
        """Test case for slices_delete_slice_id_delete

        Delete slice.
        """
        response = self.client.open(
            '//slices/delete/{slice_id}'.format(slice_id='slice_id_example'),
            method='DELETE')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_get(self):
        """Test case for slices_get

        Retrieve a listing of user slices
        """
        query_string = [('name', 'name_example'),
                        ('as_self', true),
                        ('states', 'states_example'),
                        ('limit', 200),
                        ('offset', 1)]
        response = self.client.open(
            '//slices',
            method='GET',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_modify_slice_id_accept_post(self):
        """Test case for slices_modify_slice_id_accept_post

        Accept the last modify an existing slice
        """
        response = self.client.open(
            '//slices/modify/{slice_id}/accept'.format(slice_id='slice_id_example'),
            method='POST')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_modify_slice_id_put(self):
        """Test case for slices_modify_slice_id_put

        Modify an existing slice
        """
        body = 'body_example'
        response = self.client.open(
            '//slices/modify/{slice_id}'.format(slice_id='slice_id_example'),
            method='PUT',
            data=json.dumps(body),
            content_type='text/plain')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_renew_slice_id_post(self):
        """Test case for slices_renew_slice_id_post

        Renew slice
        """
        query_string = [('lease_end_time', 'lease_end_time_example')]
        response = self.client.open(
            '//slices/renew/{slice_id}'.format(slice_id='slice_id_example'),
            method='POST',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_slice_id_get(self):
        """Test case for slices_slice_id_get

        slice properties
        """
        query_string = [('as_self', true),
                        ('graph_format', 'GRAPHML')]
        response = self.client.open(
            '//slices/{slice_id}'.format(slice_id='slice_id_example'),
            method='GET',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))


if __name__ == '__main__':
    import unittest
    unittest.main()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.kernel_tick import KernelTick


class TickTest(unittest.TestCase):
    from fabric_cf.actor.core.container.globals import Globals
    Globals.config_file = "./config/config.test.yaml"
    Constants.SUPERBLOCK_LOCATION = './state_recovery.lock'

    from fabric_cf.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().initialize()

    def get_tick(self):
        return KernelTick()

    def test_a_create(self):
        tick = self.get_tick()
        tick.initialize()

        self.assertEqual(0, tick.get_beginning_of_time())
        self.assertEqual(1, tick.get_cycle_millis())
        self.assertEqual(False, tick.is_manual())
        self.assertEqual(True, tick.stopped)
        self.assertIsNotNone(tick.clock)
        self.assertIsNotNone(tick.logger)
        self.assertIsNotNone(tick.subscribers)

    def test_b_properties(self):
        beginning = 1000
        cycle_length = 234

        tick = self.get_tick()
        tick.set_beginning_of_time(value=beginning)
        tick.set_cycle_millis(cycle_millis=cycle_length)
        tick.initialize()

        self.assertEqual(beginning, tick.get_beginning_of_time())
        self.assertEqual(cycle_length, tick.get_cycle_millis())

        failed = False

        try:
            tick.set_cycle_millis(cycle_millis=cycle_length + 10)
        except Exception:
            failed = True

        if not failed:
            self.fail()

        self.assertEqual(beginning, tick.get_beginning_of_time())

    def test_c_start_stop(self):
        tick = self.get_tick()
        tick.initialize()

        failed = False

        try:
            tick.tick()
        except Exception:
            failed = True

        self.assertTrue(failed)

        tick.start()

        failed = False

        try:
            tick.start()
        except Exception:
            failed = True

        self.assertTrue(failed)

        tick.stop()

        failed = False

        try:
            tick.stop()
        except Exception:
            failed = True

        self.assertTrue(failed)

        tick.start()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from http.client import BAD_REQUEST, UNAUTHORIZED, FORBIDDEN, NOT_FOUND
from typing import Union

import connexion

from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.swagger_server.response.cors_response import cors_400, cors_401, cors_403, cors_404, \
    cors_response, cors_500, cors_200


def get_token() -> str:
    result = None
    token = connexion.request.headers.get('Authorization', None)
    if token is not None:
        token = token.replace('Bearer ', '')
        result = f"{token}"
    return result


def cors_error_response(error: Union[OrchestratorException, Exception]) -> cors_response:
    if isinstance(error, OrchestratorException):
        if error.get_http_error_code() == BAD_REQUEST:
            return cors_400(details=str(error))
        elif error.get_http_error_code() == UNAUTHORIZED:
            return cors_401(details=str(error))
        elif error.get_http_error_code() == FORBIDDEN:
            return cors_403(details=str(error))
        elif error.get_http_error_code() == NOT_FOUND:
            return cors_404(details=str(error))
        else:
            return cors_500(details=str(error))
    else:
        return cors_500(details=str(error))


def cors_success_response(response_body) -> cors_response:
    return cors_200(response_body=response_body)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading

from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
from fabric_cf.actor.core.apis.abc_proxy import ABCProxy
from fabric_cf.actor.core.common.exceptions import RegistryException
from fabric_cf.actor.core.registry.callback_registry import CallbackRegistry
from fabric_cf.actor.core.registry.proxy_registry import ProxyRegistry


class ActorRegistry:
    """
    Maintains a collection of uniquely named actors residing in a given docker container or
    known to the docker container (singleton), indexed by type and by owner.
    getActor(String) now looks up EITHER by name or guid - tries one, then the other
    """
    class ActorRegistryEntry:
        def __init__(self, *, actor: ABCActorMixin):
            self.actor = actor
            self.kafka_topics = {}

        def get_actor(self) -> ABCActorMixin:
            return self.actor

        def get_kafka_topic(self, *, protocol: str) -> str:
            return self.kafka_topics[protocol]

        def register_kafka_topic(self, *, protocol: str, kafka_topic: str):
            self.kafka_topics[protocol] = kafka_topic

    def __init__(self):
        self.actors = {}
        self.actors_by_guid = {}
        self.proxies = ProxyRegistry()
        self.callbacks = CallbackRegistry()
        self.lock = threading.Lock()

    def clear(self):
        try:
            self.lock.acquire()
            self.actors.clear()
            self.actors_by_guid.clear()
            self.proxies.clear()
            self.callbacks.clear()
        finally:
            self.lock.release()

    def get_actor(self, *, actor_name_or_guid: str) -> ABCActorMixin:
        result = None
        entry = None
        try:
            self.lock.acquire()
            if actor_name_or_guid in self.actors:
                entry = self.actors[actor_name_or_guid]

            if actor_name_or_guid in self.actors_by_guid:
                entry = self.actors_by_guid[actor_name_or_guid]

            if entry is not None:
                result = entry.actor
        finally:
            self.lock.release()
        return result

    def get_actors(self) -> list:
        result = []
        try:
            self.lock.acquire()
            for actor in self.actors.values():
                result.append(actor.get_actor())
        finally:
            self.lock.release()
        return result

    def get_broker_proxies(self, *, protocol: str) -> list:
        result = None
        try:
            self.lock.acquire()
            result = self.proxies.get_broker_proxies(protocol=protocol)
        finally:
            self.lock.release()
        return result

    def get_callback(self, *, protocol: str, actor_name: str) -> ABCCallbackProxy:
        result = None
        try:
            self.lock.acquire()
            result = self.callbacks.get_callback(protocol=protocol, actor_name=actor_name)
        finally:
            self.lock.release()
        return result

    def get_kafka_topic(self, *, protocol: str, actor_name: str):
        result = None
        try:
            self.lock.acquire()
            if actor_name in self.actors:
                result = self.actors[actor_name].get_kafka_topic(protocol)
        finally:
            self.lock.release()
        return result

    def get_proxies(self, *, protocol: str) -> list:
        result = None
        try:
            self.lock.acquire()
            result = self.proxies.get_proxies(protocol=protocol)
        finally:
            self.lock.release()
        return result

    def get_proxy(self, *, protocol: str, actor_name: str) -> ABCProxy:
        result = None
        try:
            self.lock.acquire()
            result = self.proxies.get_proxy(protocol=protocol, actor_name=actor_name)
        finally:
            self.lock.release()
        return result

    def get_site_proxies(self, *, protocol: str) -> list:
        result = None
        try:
            self.lock.acquire()
            result = self.proxies.get_site_proxies(protocol=protocol)
        finally:
            self.lock.release()
        return result

    def register_actor(self, *, actor: ABCActorMixin):
        actor_name = actor.get_name()
        guid = actor.get_guid()

        try:
            self.lock.acquire()
            if actor_name in self.actors or guid in self.actors_by_guid:
                raise RegistryException("An actor name {} already exists".format(actor_name))

            entry = self.ActorRegistryEntry(actor=actor)
            self.actors_by_guid[guid] = entry
            self.actors[actor_name] = entry
        finally:
            self.lock.release()

    def register_callback(self, *, callback: ABCCallbackProxy):
        try:
            self.lock.acquire()
            self.callbacks.register_callback(callback=callback)
        finally:
            self.lock.release()

    def register_kafka_topic(self, *, actor_name: str, protocol: str, kafka_topic: str):
        try:
            self.lock.acquire()
            if actor_name not in self.actors:
                raise RegistryException("Actor name {} not found".format(actor_name))

            entry = self.actors[actor_name]
            entry.register_kafka_topic(protocol, kafka_topic)
        finally:
            self.lock.release()

    def register_proxy(self, *, proxy: ABCProxy):
        try:
            self.lock.acquire()
            self.proxies.register_proxy(proxy=proxy)
        finally:
            self.lock.release()

    def unregister(self, *, actor: ABCActorMixin):
        try:
            self.lock.acquire()
            actor_name = actor.get_name()
            guid = actor.get_guid()

            if actor_name in self.actors:
                self.actors.pop(actor_name)
            if guid in self.actors_by_guid:
                self.actors_by_guid.pop(guid)
            self.callbacks.unregister(actor_name=actor_name)
            self.proxies.unregister(actor_name=actor_name)
        finally:
            self.lock.release()


class ActorRegistrySingleton:
    __instance = None

    def __init__(self):
        if self.__instance is not None:
            raise RegistryException("Singleton can't be created twice !")

    def get(self):
        """
        Actually create an instance
        """
        if self.__instance is None:
            self.__instance = ActorRegistry()
        return self.__instance

    get = classmethod(get)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from fabric_mb.message_bus.messages.auth_avro import AuthAvro
from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.broker_query_model_avro import BrokerQueryModelAvro
from fabric_mb.message_bus.messages.poa_avro import PoaAvro
from fabric_mb.message_bus.messages.poa_info_avro import PoaInfoAvro
from fabric_mb.message_bus.messages.resource_ticket_avro import ResourceTicketAvro
from fabric_mb.message_bus.messages.resource_set_avro import ResourceSetAvro
from fabric_mb.message_bus.messages.result_avro import ResultAvro
from fabric_mb.message_bus.messages.result_poa_avro import ResultPoaAvro
from fabric_mb.message_bus.messages.site_avro import SiteAvro
from fabric_mb.message_bus.messages.slice_avro import SliceAvro
from fabric_mb.message_bus.messages.term_avro import TermAvro
from fabric_mb.message_bus.messages.unit_avro import UnitAvro
from fabric_mb.message_bus.messages.update_data_avro import UpdateDataAvro
from fabric_mb.message_bus.messages.ticket import Ticket as AvroTicket

from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import ProxyException
from fabric_cf.actor.core.core.ticket import Ticket
from fabric_cf.actor.core.core.unit import Unit, UnitState
from fabric_cf.actor.core.core.unit_set import UnitSet
from fabric_cf.actor.core.delegation.resource_ticket import ResourceTicket
from fabric_cf.actor.core.kernel.poa import Poa, PoaFactory
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.kernel.slice import SliceFactory
from fabric_cf.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric_cf.actor.core.container.maintenance import Site
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.resource_type import ResourceType
from fabric_cf.actor.core.util.update_data import UpdateData
from fabric_cf.actor.db import Units
from fabric_cf.actor.security.auth_token import AuthToken


if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice


class Translate:
    @staticmethod
    def translate_udd(*, udd: UpdateData) -> UpdateDataAvro:
        result = UpdateDataAvro()
        result.message = udd.get_message()
        result.failed = udd.is_failed()
        return result

    @staticmethod
    def translate_udd_from_avro(*, udd: UpdateDataAvro) -> UpdateData:
        result = UpdateData()
        result.message = udd.message
        result.failed = udd.failed
        return result

    @staticmethod
    def translate_slice(*, slice_avro: SliceAvro) -> ABCSlice:
        if slice_avro is None:
            return None
        if slice_avro.guid is None and slice_avro.slice_name is None:
            return None

        if slice_avro.guid is None:
            raise ProxyException(Constants.NOT_SPECIFIED_PREFIX.format("Slice id"))

        slice_obj = SliceFactory.create(slice_id=ID(uid=slice_avro.guid),
                                        name=slice_avro.slice_name,
                                        project_id=slice_avro.project_id,
                                        project_name=slice_avro.project_name)
        slice_obj.set_description(description=slice_avro.description)
        slice_obj.set_config_properties(value=slice_avro.config_properties)
        slice_obj.set_lease_start(lease_start=slice_avro.get_lease_start())
        slice_obj.set_lease_end(lease_end=slice_avro.get_lease_end())
        slice_obj.set_graph_id(graph_id=slice_avro.graph_id)
        return slice_obj

    @staticmethod
    def translate_slice_to_avro(*, slice_obj: ABCSlice) -> SliceAvro:
        avro_slice = SliceAvro()
        avro_slice.slice_name = slice_obj.get_name()
        avro_slice.guid = str(slice_obj.get_slice_id())
        avro_slice.description = slice_obj.get_description()
        avro_slice.owner = Translate.translate_auth_to_avro(auth=slice_obj.get_owner())
        avro_slice.state = slice_obj.get_state().value
        avro_slice.config_properties = slice_obj.get_config_properties()
        avro_slice.set_lease_end(lease_end=slice_obj.get_lease_end())
        avro_slice.set_lease_start(lease_start=slice_obj.get_lease_start())
        if hasattr(slice_obj, 'project_id'):
            avro_slice.set_project_id(project_id=slice_obj.get_project_id())

        if hasattr(slice_obj, 'project_name'):
            avro_slice.set_project_name(project_name=slice_obj.get_project_name())

        if slice_obj.get_resource_type() is not None:
            avro_slice.set_resource_type(str(slice_obj.get_resource_type()))

        avro_slice.set_client_slice(slice_obj.is_client())

        if slice_obj.get_graph_id() is not None:
            avro_slice.graph_id = slice_obj.get_graph_id()

        return avro_slice

    @staticmethod
    def translate_site_to_avro(*, site: Site) -> SiteAvro or None:
        if site is None:
            return None
        result = SiteAvro()
        result.name = site.get_name()
        result.maint_info = site.get_maintenance_info()
        return result

    @staticmethod
    def translate_site_from_avro(*, site_avro: SiteAvro) -> Site or None:
        if site_avro is None:
            return None
        result = Site(name=site_avro.get_name(), maint_info=site_avro.get_maint_info())
        return result

    @staticmethod
    def translate_auth_to_avro(*, auth: AuthToken) -> AuthAvro:
        result = AuthAvro()
        result.name = auth.get_name()
        result.guid = str(auth.get_guid())
        result.oidc_sub_claim = auth.get_oidc_sub_claim()
        result.email = auth.get_email()
        return result

    @staticmethod
    def translate_auth_from_avro(*, auth_avro: AuthAvro) -> AuthToken:
        if auth_avro is None:
            return None
        result = AuthToken()
        result.name = auth_avro.name
        result.guid = auth_avro.guid
        result.oidc_sub_claim = auth_avro.oidc_sub_claim
        result.email = auth_avro.email
        return result

    @staticmethod
    def translate_term(*, term: Term) -> TermAvro:
        avro_term = TermAvro()
        if term.get_start_time() is not None:
            avro_term.start_time = ActorClock.to_milliseconds(when=term.get_start_time())
        else:
            term.start_time = 0
        if term.get_end_time() is not None:
            avro_term.end_time = ActorClock.to_milliseconds(when=term.get_end_time())
        else:
            term.end_time = 0
        if term.get_new_start_time() is not None:
            avro_term.new_start_time = ActorClock.to_milliseconds(when=term.get_new_start_time())
        else:
            avro_term.new_start_time = 0
        return avro_term

    @staticmethod
    def translate_resource_ticket(*, resource_ticket: ResourceTicket) -> ResourceTicketAvro:
        rd = ResourceTicketAvro()
        rd.units = resource_ticket.units
        if resource_ticket.term is not None:
            rd.term = Translate.translate_term(term=resource_ticket.term)

        if resource_ticket.type is not None:
            rd.type = str(resource_ticket.type)

        if resource_ticket.guid is not None:
            rd.guid = str(resource_ticket.guid)

        if resource_ticket.properties is not None:
            rd.properties = resource_ticket.properties

        if resource_ticket.issuer is not None:
            rd.issuer = str(resource_ticket.issuer)

        if resource_ticket.holder is not None:
            rd.holder = str(resource_ticket.holder)

        return rd

    @staticmethod
    def translate_resource_ticket_from_avro(*, resource_ticket: ResourceTicketAvro) -> ResourceTicket:
        rd = ResourceTicket()
        rd.units = resource_ticket.units
        if resource_ticket.term is not None:
            rd.term = Translate.translate_term_from_avro(term=resource_ticket.term)

        if resource_ticket.type is not None:
            rd.type = ResourceType(resource_type=resource_ticket.type)

        if resource_ticket.guid is not None:
            rd.guid = ID(uid=resource_ticket.guid)

        if resource_ticket.properties is not None:
            rd.properties = resource_ticket.properties

        if resource_ticket.issuer is not None:
            rd.issuer = ID(uid=resource_ticket.issuer)

        if resource_ticket.holder is not None:
            rd.holder = ID(uid=resource_ticket.holder)

        return rd

    @staticmethod
    def translate_ticket(*, ticket: Ticket) -> AvroTicket:
        avro_ticket = AvroTicket()
        avro_ticket.authority = Translate.translate_auth_to_avro(auth=ticket.authority.get_identity())
        avro_ticket.old_units = ticket.old_units
        avro_ticket.delegation_id = ticket.delegation_id
        avro_ticket.resource_ticket = Translate.translate_resource_ticket(
            resource_ticket=ticket.resource_ticket)
        return avro_ticket

    @staticmethod
    def translate_ticket_from_avro(*, avro_ticket: AvroTicket) -> Ticket:
        ticket = Ticket()
        ticket.delegation_id = avro_ticket.delegation_id
        auth_identity = Translate.translate_auth_from_avro(auth_avro=avro_ticket.authority)
        ticket.authority = ActorRegistrySingleton.get().get_proxy(protocol=Constants.PROTOCOL_KAFKA,
                                                                  actor_name=auth_identity.get_name())
        ticket.old_units = avro_ticket.old_units
        if avro_ticket.resource_ticket is not None:
            ticket.resource_ticket = Translate.translate_resource_ticket_from_avro(
                resource_ticket=avro_ticket.resource_ticket)

        return ticket

    @staticmethod
    def translate_resource_set(*, resource_set: ResourceSet) -> ResourceSetAvro:
        avro_rset = ResourceSetAvro()
        avro_rset.type = str(resource_set.get_type())
        avro_rset.units = resource_set.get_units()
        avro_rset.set_sliver(sliver=resource_set.get_sliver())
        return avro_rset

    @staticmethod
    def translate_term_from_avro(*, term: TermAvro) -> Term:
        start_time = None
        end_time = None
        new_start_time = None
        if term.start_time > 0:
            start_time = ActorClock.from_milliseconds(milli_seconds=term.start_time)

        if term.end_time > 0:
            end_time = ActorClock.from_milliseconds(milli_seconds=term.end_time)

        if term.new_start_time > 0:
            new_start_time = ActorClock.from_milliseconds(milli_seconds=term.new_start_time)

        return Term(start=start_time, end=end_time, new_start=new_start_time)

    @staticmethod
    def translate_resource_set_from_avro(*, rset: ResourceSetAvro) -> ResourceSet:
        result = ResourceSet(units=rset.units, rtype=ResourceType(resource_type=rset.type), sliver=rset.get_sliver())
        return result

    @staticmethod
    def attach_properties(*, slice_mng: SliceAvro, slice_obj: ABCSlice) -> SliceAvro:
        if slice_obj.get_config_properties() is not None:
            slice_mng.set_config_properties(slice_obj.get_config_properties())

        return slice_mng

    @staticmethod
    def fill_slices(*, slice_list: List[ABCSlice], full: bool) -> List[SliceAvro]:
        result = []
        for slice_obj in slice_list:
            if slice_obj is not None:
                ss = Translate.translate_slice_to_avro(slice_obj=slice_obj)
                if full:
                    ss = Translate.attach_properties(slice_mng=ss, slice_obj=slice_obj)
                result.append(ss)
        return result

    @staticmethod
    def translate_delegation_to_avro(*, delegation: ABCDelegation) -> DelegationAvro:
        avro_delegation = DelegationAvro()
        avro_delegation.delegation_id = delegation.get_delegation_id()
        avro_delegation.state = delegation.get_state().value
        avro_delegation.slice = Translate.translate_slice_to_avro(slice_obj=delegation.get_slice_object())
        avro_delegation.site = delegation.get_site()
        if delegation.get_graph() is not None and not (delegation.is_reclaimed() or delegation.is_closed()):
            avro_delegation.graph = delegation.get_graph().serialize_graph()
        return avro_delegation

    @staticmethod
    def translate_to_broker_query_model(*, query_response: dict, level: int) -> BrokerQueryModelAvro:
        bqm = BrokerQueryModelAvro()
        bqm.level = level
        bqm.model = query_response.get(Constants.BROKER_QUERY_MODEL, None)
        return bqm

    @staticmethod
    def translate_unit_set(*, unit_set: UnitSet) -> List[Units]:
        result = None
        if unit_set.units is not None:
            result = []
            for u in unit_set.units.values():
                result.append(Translate.translate_unit(unit=u))
        return result

    @staticmethod
    def translate_unit_from_avro(*, unit_avro: UnitAvro) -> Unit:
        unit = Unit(rid=ID(uid=unit_avro.reservation_id))

        if unit_avro.properties is not None:
            unit.properties = unit_avro.properties

        if unit_avro.rtype is not None:
            unit.rtype = ResourceType(resource_type=unit.rtype)

        if unit_avro.parent_id is not None:
            unit.parent_id = ID(uid=unit_avro.parent_id)

        if unit_avro.slice_id is not None:
            unit.slice_id = ID(uid=unit_avro.slice_id)

        if unit_avro.actor_id is not None:
            unit.actor_id = ID(uid=unit_avro.actor_id)

        if unit_avro.state is not None:
            unit.state = UnitState(unit_avro.state)
        return unit

    @staticmethod
    def translate_unit(*, unit: Unit) -> UnitAvro:
        result = UnitAvro()
        result.properties = unit.properties
        if unit.rtype is not None:
            result.rtype = str(unit.rtype)

        if unit.parent_id is not None:
            result.parent_id = str(unit.parent_id)

        if unit.reservation_id is not None:
            result.reservation_id = str(unit.reservation_id)

        if unit.slice_id is not None:
            result.slice_id = str(unit.slice_id)

        if unit.actor_id is not None:
            result.actor_id = str(unit.actor_id)

        if unit.state is not None:
            result.state = unit.state.value
        return result

    @staticmethod
    def translate_unit_set_from_avro(*, unit_list: List[UnitAvro]) -> UnitSet:
        unit_set = UnitSet(plugin=None)
        unit_set.units = {}
        for u in unit_list:
           obj = Translate.translate_unit_from_avro(unit_avro=u)
           unit_set.units[obj.get_id()] = obj
        return unit_set

    @staticmethod
    def translate_poa_avro_to_poa(*, poa_avro: PoaAvro) -> Union[Poa, None]:
        if poa_avro is None:
            return poa_avro
        if poa_avro.poa_id is None and poa_avro.rid is None:
            return None

        poa_obj = PoaFactory.create(poa_id=poa_avro.poa_id, operation=poa_avro.get_operation(),
                                    sliver_id=ID(uid=poa_avro.rid), vcpu_cpu_map=poa_avro.get_vcpu_cpu_map(),
                                    node_set=poa_avro.get_node_set())
        poa_obj.sequence_poa_in = poa_avro.sequence
        poa_obj.slice_id = poa_avro.slice_id
        # NOTE: Sliver, Slice and Project Info to be set by the caller
        return poa_obj

    @staticmethod
    def translate_poa_to_poa_avro(*, poa: Poa) -> Union[PoaAvro, None]:
        if poa is None:
            return poa

        auth_avro = Translate.translate_auth_to_avro(auth=poa.get_slice().get_owner())
        poa_avro = PoaAvro(operation=poa.operation, rid=str(poa.sliver_id), poa_id=poa.poa_id,
                           vcpu_cpu_map=poa.vcpu_cpu_map, node_set=poa.node_set, auth=auth_avro,
                           project_id=poa.get_slice().get_project_id(), slice_id=str(poa.get_slice_id()))
        poa_avro.sequence = poa.sequence_poa_out
        return poa_avro

    @staticmethod
    def translate_poa_to_poa_info_avro(*, poa: Poa) -> Union[PoaInfoAvro, None]:
        if poa is None:
            return poa

        auth_avro = Translate.translate_auth_to_avro(auth=poa.get_slice().get_owner())
        poa_avro = PoaInfoAvro(operation=poa.operation, rid=str(poa.sliver_id), poa_id=str(poa.poa_id),
                               auth=auth_avro, project_id=poa.get_slice().get_project_id(),
                               slice_id=str(poa.get_slice().get_slice_id()), state=str(poa.get_state()),
                               error=poa.get_message())

        if poa.get_info() is not None:
            poa_avro.info = poa.get_info().copy()
        return poa_avro

    @staticmethod
    def translate_poa_to_result_poa_avro(*, poa: Poa) -> Union[ResultPoaAvro, None]:
        if poa is None:
            return poa

        poa_info_avro = Translate.translate_poa_to_poa_info_avro(poa=poa)

        poa_result_avro = ResultPoaAvro()
        poa_result_avro.status = ResultAvro()
        poa_result_avro.status.code = poa.get_error_code()
        poa_result_avro.status.message = poa.get_message()
        poa_result_avro.poas = [poa_info_avro]

        return poa_result_avro

    @staticmethod
    def translate_result_poa_avro_to_poa(*, poa_result: ResultPoaAvro) -> Union[Poa, None]:
        if poa_result is None:
            return poa_result
        if poa_result.poas is None or len(poa_result.poas) <= 0:
            return None

        poa_info = poa_result.poas[0]

        poa_obj = PoaFactory.create(poa_id=poa_info.poa_id, operation=poa_info.get_operation(),
                                    sliver_id=ID(uid=poa_info.rid))
        poa_obj.slice_id = poa_info.get_slice_id()
        poa_obj.error_code = poa_result.get_status().get_code()
        poa_obj.message = poa_result.get_status().get_message()
        poa_obj.info = poa_info.get_info()

        return poa_obj
